# -*- coding: utf-8 -*-
import logging
import math
import os

import numpy as np
import pandas as pd
import sklearn
from sklearn import tree
from sklearn.neural_network import MLPClassifier

from ._helper import _get_rounding_decimal_places
from ._input_generator import _InputGenerator
from ._leo_helper import _get_leo_integer_type


def _get_model_transpiler(model, validation_data, fixed_point_scaling_factor=None, **kwargs):
    if isinstance(model, sklearn.tree._classes.DecisionTreeClassifier):
        return _DecisionTreeTranspiler(model, validation_data, fixed_point_scaling_factor)
    elif isinstance(
        model, sklearn.neural_network._multilayer_perceptron.MLPClassifier
    ) or isinstance(model, sklearn.neural_network._multilayer_perceptron.MLPRegressor):
        # ensure that the model uses the ReLU activation function
        if model.activation != "relu":
            raise ValueError(
                "The model uses the activation function "
                f"{model.activation}, but only ReLU is supported."
            )
        return _MLPTranspiler(model, validation_data, fixed_point_scaling_factor, **kwargs)
    else:
        raise ValueError("Model is not supported.")


class _ModelTranspilerBase:
    def __init__(self, model, validation_data, pre_set_fixed_point_scaling_factor):
        self.model = model
        self.validation_data = validation_data
        self.pre_set_fixed_point_scaling_factor = pre_set_fixed_point_scaling_factor

    def transpile(self, project_name, model_as_input):
        raise NotImplementedError("This method is not implemented.")

    def _numbers_get_leo_type_and_fixed_point_scaling_factor(self):
        minimum_model, maximum_model = self._get_numeric_range_model()
        minimum = minimum_model
        maximum = maximum_model
        if self.validation_data is not None:
            minimum_data, maximum_data = self._get_numeric_range_data()
            minimum = min(minimum, minimum_data)
            maximum = max(maximum, maximum_data)

            (
                minimum_model_inference,
                maximum_model_inference,
            ) = self._get_numeric_range_model_inference()

            if minimum_model_inference is not None:
                minimum = min(minimum, minimum_model_inference)
            if maximum_model_inference is not None:
                maximum = max(maximum, maximum_model_inference)

        bits_for_integer_part = math.ceil(math.log2(max(abs(minimum), abs(maximum))))
        signed_type_needed = minimum < 0

        # Fixed point parametrization
        # todo implement this for model inference
        max_decimal_places_model = self._get_max_decimal_places_model()
        max_decimal_places_data = 0
        if self.validation_data is not None:
            max_decimal_places_data = self._get_max_decimal_places_data()
        max_decimal_places = max(max_decimal_places_data, max_decimal_places_model)

        min_decimal_value = 10 ** (-max_decimal_places)
        fixed_point_min_scaling_exponent = math.log2(1 / min_decimal_value)
        bits_for_fractional_part = math.ceil(fixed_point_min_scaling_exponent)

        if(self.pre_set_fixed_point_scaling_factor is None):
            fixed_point_scaling_factor = 2**bits_for_fractional_part
        else:
            fixed_point_scaling_factor = self.pre_set_fixed_point_scaling_factor

        if self.validation_data is not None:
            (
                minimum_model_inference_fixed_point,
                maximum_model_inference_fixed_point,
            ) = self._get_numeric_range_model_inference(
                scaling_factor=fixed_point_scaling_factor
            )

            if minimum_model_inference is not None:
                bits_for_integer_part_after_fixed_point_inference = math.ceil(
                    math.log2(
                        max(
                            abs(minimum_model_inference_fixed_point),
                            abs(maximum_model_inference_fixed_point),
                        )
                    )
                )
                bits_for_integer_part = max(
                    bits_for_integer_part,
                    bits_for_integer_part_after_fixed_point_inference,
                )

        leo_type = _get_leo_integer_type(
            signed_type_needed, bits_for_integer_part + bits_for_fractional_part
        )

        self.leo_type = leo_type
        self.fixed_point_scaling_factor = fixed_point_scaling_factor
        self.output_fixed_point_scaling_factor_power = 1

        logging.info(
            f"Minimum number: {minimum}, maximum number: {maximum}. Recommended "
            f"fixed-point scaling factor: {fixed_point_scaling_factor}, required Leo "
            f"type: {leo_type}"
        )

        return leo_type, fixed_point_scaling_factor

    def _get_numeric_range_model(self):
        raise NotImplementedError("This method is not implemented.")

    def _get_numeric_range_data(self):
        if isinstance(self.validation_data, np.ndarray):
            return self.validation_data.min(), self.validation_data.max()
        elif isinstance(self.validation_data, pd.DataFrame):
            return self.validation_data.min().min(), self.validation_data.max().max()
        else:
            raise TypeError("Unsupported data type for validation_data")

    def _get_numeric_range_model_inference(self, scaling_factor=1):
        # has to be implemented by the subclass if the model requires it
        return None, None

    def _get_max_decimal_places_model(self):
        raise NotImplementedError("This method is not implemented.")

    def _get_max_decimal_places_data(self):
        if isinstance(self.validation_data, np.ndarray):
            return max(
                [
                    _get_rounding_decimal_places(val)
                    for val in self.validation_data.ravel()
                ]
            )
        elif isinstance(self.validation_data, pd.DataFrame):
            return max(
                [
                    _get_rounding_decimal_places(val)
                    for val in self.validation_data.to_numpy().ravel()
                ]
            )
        else:
            raise TypeError("Unsupported data type for validation_data")

    def _convert_to_fixed_point(self, value, power=1):
        if hasattr(value, "shape"):  # check if value is a numpy array
            vectorized_int = np.vectorize(int)
            return vectorized_int(
                value.astype(object) * (self.fixed_point_scaling_factor**power)
            )
        else:
            return int(round(value * (self.fixed_point_scaling_factor**power)))

    def convert_computation_base_outputs_to_decimal(self, computation_base):
        computation_base.fixed_point_scaling_factor = (
            self.fixed_point_scaling_factor
            ** self.output_fixed_point_scaling_factor_power
        )
        computation_base.output_decimal = self._convert_from_fixed_point(
            computation_base.output
        )

    def _convert_from_fixed_point(self, value):
        if isinstance(value, list):
            return [self._convert_from_fixed_point(val) for val in value]
        else:
            return value / (
                self.fixed_point_scaling_factor
                ** self.output_fixed_point_scaling_factor_power
            )

    def _get_fixed_point_and_leo_type(self, value):
        return str(self._convert_to_fixed_point(value)) + self.leo_type

    def _merge_into_transpiled_code(
        self,
        project_name,
        struct_definitions,
        circuit_inputs,
        circuit_outputs,
        model_logic_snippets,
        add_relu_function=False,
    ):
        code = ""
        code += """// This file was automatically generated by the """
        code += f"""zkml LeoTranspiler.
program {project_name}.aleo {{
    {struct_definitions}
    transition main {circuit_inputs} -> {circuit_outputs} {{\n"""

        for element in model_logic_snippets:
            if isinstance(element, str):
                code += f"""{element}"""
            elif isinstance(element, _InputGenerator._Input):
                code += f"""{element.reference_name}"""
            else:
                raise ValueError("Unknown element type in model logic snippets.")

        code += """    }"""

        if add_relu_function and False:  # todo adjust to data type
            code += """
                function relu(x: field) -> field {
        let x_integer: i128 = x as i128;
        if x_integer < 0i128 {
            return 0field;
        } else {
            return x_integer as field;
        }
    }"""

        if add_relu_function and True:  # todo adjust to data type
            code += f"function relu(x: {self.leo_type}) -> {self.leo_type} {{\n"
            code += f"if x < 0{self.leo_type} {{"
            code += f"return 0{self.leo_type};"
            code += "} else {"
            code += "return x;"
            code += """
        }
    }"""
        code += """
}"""
        return code

    def generate_input(self, features):
        fixed_point_features = self._convert_to_fixed_point(features)
        return self.input_generator.generate_input(fixed_point_features)


class _DecisionTreeTranspiler(_ModelTranspilerBase):
    def __init__(self, model, validation_data, pre_set_fixed_point_scaling_factor):
        super().__init__(model, validation_data, pre_set_fixed_point_scaling_factor)

    def transpile(self, project_name: str, model_as_input: bool):
        """
        Transpile a model to Leo.

        Parameters
        ----------
        project_name : str
            The name of the project.
        model_as_input : bool
            Whether the model parameters should be an input to the circuit.

        Returns
        -------
        transpilation_result : str
            The transpiled model.
        """
        tree = self.model.tree_

        # Input generation
        self.input_generator = _InputGenerator()
        for _ in range(tree.n_features):
            self.input_generator.add_input(self.leo_type, "xi")

        decision_tree_logic_snippets = self._transpile_decision_tree_logic_to_leo_code(
            tree, model_as_input, indentation="        "
        )

        (
            struct_definitions,
            input_string,
            active_input_count,
        ) = self.input_generator.get_struct_definitions_and_circuit_input_string()
        circuit_inputs = "(" + input_string + ")"
        circuit_outputs = (
            "(" + self.leo_type + ")"
        )  # Todo check multi output decision trees and models

        transpilation_result = self._merge_into_transpiled_code(
            project_name,
            struct_definitions,
            circuit_inputs,
            circuit_outputs,
            decision_tree_logic_snippets,
        )
        self.active_input_count = active_input_count
        return transpilation_result

    def _transpile_decision_tree_logic_to_leo_code(
        self, tree, model_as_input, node=0, indentation=""
    ):
        left_child = tree.children_left[node]
        right_child = tree.children_right[node]

        # Base case: leaf node
        if left_child == right_child:  # means it's a leaf
            return [
                indentation + f"return "
                f"{self._get_fixed_point_and_leo_type(tree.value[node].argmax())};\n"
            ]

        # Recursive case: internal node
        feature = self.input_generator.use_input(tree.feature[node])
        threshold = self._convert_to_fixed_point(tree.threshold[node])

        leo_code_snippets = []

        if node == 0:
            leo_code_snippets += [indentation + "if ", feature, " <= "]
            if model_as_input:
                leo_code_snippets += [
                    self.input_generator.add_input(
                        self.leo_type, "customi", True, threshold, "threshold"
                    )
                ]
            else:
                leo_code_snippets += [f"{threshold}{self.leo_type}"]
            leo_code_snippets += [
                " {\n",
            ]
        else:
            leo_code_snippets += [indentation + "if ", feature, " <= "]
            if model_as_input:
                leo_code_snippets += [
                    self.input_generator.add_input(
                        self.leo_type, "customi", True, threshold, "threshold"
                    )
                ]
            else:
                leo_code_snippets += [f"{threshold}{self.leo_type}"]
            leo_code_snippets += [
                " {\n",
            ]

        leo_code_snippets += self._transpile_decision_tree_logic_to_leo_code(
            tree, model_as_input, left_child, indentation + "    "
        )
        leo_code_snippets += [indentation + "}\n" + indentation + "else {\n"]

        leo_code_snippets += self._transpile_decision_tree_logic_to_leo_code(
            tree, model_as_input, right_child, indentation + "    "
        )
        leo_code_snippets += [indentation + "}\n"]
        return leo_code_snippets

    def _get_numeric_range_model(self):
        thresholds = self.model.tree_.threshold
        minimum = min(thresholds)
        maximum = max(thresholds)

        classes = self.model.classes_
        # check if classes are numeric
        if isinstance(classes[0], int):
            minimum = min(minimum, min(classes))
            maximum = max(maximum, max(classes))

        return minimum, maximum

    def _get_max_decimal_places_model(self):
        max_decimal_places = max(
            [
                _get_rounding_decimal_places(val)
                for val in self.model.tree_.threshold.ravel()
            ]
        )
        return max_decimal_places


class _MLPTranspiler(_ModelTranspilerBase):
    def __init__(self, model, validation_data, pre_set_fixed_point_scaling_factor, **kwargs):
        super().__init__(model, validation_data, pre_set_fixed_point_scaling_factor)

        if("data_representation_type" in kwargs):
            self.data_representation_type = kwargs["data_representation_type"]
        else:
            self.data_representation_type = "int"

        if("layer_wise_fixed_point_scaling_factor" in kwargs):
            self.layer_wise_fixed_point_scaling_factor = kwargs["layer_wise_fixed_point_scaling_factor"]
        else:
            self.layer_wise_fixed_point_scaling_factor = True

    def _get_numeric_range_model(self):
        minimum = None
        maximum = None

        for layer in self.model.coefs_:
            layer_minimum = layer.min()
            layer_maximum = layer.max()
            if minimum is None:
                minimum = layer_minimum
                maximum = layer_maximum
            else:
                minimum = min(minimum, layer_minimum)
                maximum = max(maximum, layer_maximum)

        for layer in self.model.intercepts_:
            layer_minimum = layer.min()
            layer_maximum = layer.max()
            minimum = min(minimum, layer_minimum)
            maximum = max(maximum, layer_maximum)

        # check if model is a classifier
        if isinstance(self.model, MLPClassifier):
            classes = self.model.classes_
            # check if classes are numeric
            # Todo generalize this code with others like the decision trees
            # Todo one could quantize here and make the range smaller
            if isinstance(classes[0], int) or isinstance(classes[0], np.int64):
                minimum = min(minimum, min(classes))
                maximum = max(maximum, max(classes))

        return minimum, maximum

    def _get_numeric_range_model_inference(self, scaling_factor=1):
        minimum = None
        maximum = None

        for data_point in self.validation_data:
            min_inference, max_inference = self._get_min_max_pre_activation_values(
                self.model, data_point, scaling_factor
            )
            if minimum is None:
                minimum = min_inference
                maximum = max_inference
            else:
                minimum = min(minimum, min_inference)
                maximum = max(maximum, max_inference)

        return minimum, maximum

    def _get_min_max_pre_activation_values(self, model, X, scaling_factor=1):
        # todo implement rounding in fixed point number conversion
        layer_input = X
        global_min = float("inf")
        global_max = float("-inf")

        # Iterate through layers and compute weighted sum
        for i, (weights, biases) in enumerate(zip(model.coefs_, model.intercepts_)):
            pre_activation = np.dot(
                layer_input * scaling_factor, weights * scaling_factor
            ) + biases * (scaling_factor ** (i + 2))

            # Update global min and max pre-activation values and layer-specific fixed point scaled bias
            global_min = min(
                global_min,
                np.min(pre_activation),
                np.min(biases * (scaling_factor ** (i + 2))),
            )
            global_max = max(
                global_max,
                np.max(pre_activation),
                np.max(biases * (scaling_factor ** (i + 2))),
            )

            # Apply ReLU activation function except for the output layer
            if i < len(model.coefs_) - 1:
                layer_input = np.maximum(0, pre_activation)

        return global_min, global_max

    def _get_max_decimal_places_model(self):
        max_decimal_places = None
        for layer in self.model.coefs_:
            layer_max_decimal_places = max(
                [_get_rounding_decimal_places(val) for val in layer.ravel()]
            )
            if max_decimal_places is None:
                max_decimal_places = layer_max_decimal_places
            else:
                max_decimal_places = max(max_decimal_places, layer_max_decimal_places)

        for layer in self.model.intercepts_:
            layer_max_decimal_places = max(
                [_get_rounding_decimal_places(val) for val in layer.ravel()]
            )
            max_decimal_places = max(max_decimal_places, layer_max_decimal_places)
        return max_decimal_places

    def transpile(self, project_name: str, model_as_input: bool):
        """
        Transpile a model to Leo.

        Parameters
        ----------
        project_name : str
            The name of the project.
        model_as_input : bool
            Whether the model parameters should be an input to the circuit.

        Returns
        -------
        transpilation_result : str
            The transpiled model.
        """
        # Input generation
        number_of_model_inputs = self.model.coefs_[0].shape[0]
        self.input_generator = _InputGenerator()
        for _ in range(number_of_model_inputs):
            self.input_generator.add_input(self.leo_type, "xi")

        mlp_logic_snippets = self._transpile_mlp_logic_to_leo_code(
            self.model, model_as_input, indentation="        "
        )

        (
            struct_definitions,
            input_string,
            active_input_count,
        ) = self.input_generator.get_struct_definitions_and_circuit_input_string()
        circuit_inputs = "(" + input_string + ")"

        num_outputs = self.model.coefs_[-1].shape[1]
        circuit_outputs = (
            "(" + ", ".join([self.leo_type for _ in range(num_outputs)]) + ")"
        )

        transpilation_result = self._merge_into_transpiled_code(
            project_name,
            struct_definitions,
            circuit_inputs,
            circuit_outputs,
            mlp_logic_snippets,
            add_relu_function=True,
        )
        self.active_input_count = active_input_count

        if(False):
            pseudocode = self.mlp_to_pseudocode(self.model)
            with open(
                os.path.join(
                    os.getcwd(),
                    "zkml",
                    "tests",
                    "tmp",
                    "mnist",
                    project_name,
                    "pseudocode.txt",
                ),
                "w",
            ) as f:
                f.write(pseudocode)

        return transpilation_result

    def _transpile_mlp_logic_to_leo_code(
        self,
        mlp,
        model_as_input,
        indentation="",
        prune_threshold_weights=0,
        prune_threshold_bias=0,
    ):
        
        if (self.data_representation_type == "int" and self.layer_wise_fixed_point_scaling_factor):
            return self._transpile_mlp_logic_to_leo_code_3(
                mlp,
                model_as_input,
                indentation,
                prune_threshold_weights,
                prune_threshold_bias,
            )
        if(self.data_representation_type == "int" and not self.layer_wise_fixed_point_scaling_factor):
            return self._transpile_mlp_logic_to_leo_code_4(
                mlp,
                model_as_input,
                indentation,
                prune_threshold_weights,
                prune_threshold_bias,
            )
        if(self.data_representation_type == "field" and self.layer_wise_fixed_point_scaling_factor):
            return self._transpile_mlp_logic_to_leo_code_1(
                mlp,
                model_as_input,
                indentation,
                prune_threshold_weights,
                prune_threshold_bias,
            )
        if(self.data_representation_type == "field" and not self.layer_wise_fixed_point_scaling_factor):
            raise NotImplementedError("This method is not implemented. You can use the field representation with layer-wise fixed point scaling factor, or the integer representation with or without layer-wise fixed point scaling factor.")

    def _transpile_mlp_logic_to_leo_code_1(
        self,
        mlp,
        model_as_input,
        indentation="",
        prune_threshold_weights=0,
        prune_threshold_bias=0,
    ):
        # initial version
        leo_code_snippets = []

        coefs = mlp.coefs_
        intercepts = mlp.intercepts_

        # specify and convert inputs to fields
        for i in range(coefs[0].shape[0]):
            used_input = self.input_generator.use_input(i)

            leo_code_snippets.append(
                indentation + f"let {used_input.reference_name}_field: field = "
            )

            leo_code_snippets.append(used_input)

            leo_code_snippets.append(" as field;" + "\n")

            used_input.field_name = f"{used_input.reference_name}_field"

        # for each layer
        prev_neurons = [
            f"{self.input_generator.use_input(i).field_name}"
            for i in range(coefs[0].shape[0])
        ]
        for layer in range(len(coefs)):  # for each layer
            layer_code = []
            for n in range(coefs[layer].shape[1]):  # for each neuron in the layer
                terms = []
                for i in range(coefs[layer].shape[0]):  # for each input to the neuron
                    weight_input = self.input_generator.add_input(
                        self.leo_type,
                        "customi",
                        model_as_input,
                        coefs[layer][i][n],
                        f"w_{layer}_{n}_",
                    )
                    # todo adapt for case where model weights are actual inputs
                    if abs(coefs[layer][i][n]) > prune_threshold_weights:
                        terms.append(
                            f"({self._convert_to_fixed_point(weight_input.value.item())}{self.leo_type} as field)*{prev_neurons[i]}"
                        )

                if layer != len(coefs) - 1:  # if not the last layer
                    neuron_code = indentation + f"let neuron_{layer+1}_{n}: field = "
                    weights_or_bias_above_prune_thresholds = (
                        terms != [] or abs(intercepts[layer][n]) > prune_threshold_bias
                    )

                    if weights_or_bias_above_prune_thresholds:
                        neuron_code += "relu(" + " + ".join(terms)

                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            neuron_code += f" + ({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field)"

                        neuron_code += ");\n"
                    else:
                        neuron_code += "0field;\n"
                    leo_code_snippets.append(neuron_code)

                else:  # if the last layer
                    neuron_code = indentation + f"let output_{n}_field" + f" : field = "
                    self.output_fixed_point_scaling_factor_power = layer + 2

                    if terms != [] and abs(intercepts[layer][n]) > prune_threshold_bias:
                        neuron_code += f"{' + '.join(terms)}"
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            bias_input = self.input_generator.add_input(
                                self.leo_type,
                                "customi",
                                model_as_input,
                                intercepts[layer][n],
                                f"b_{layer}_{n}_",
                            )
                            neuron_code += f" + ({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field)"
                        neuron_code += ";\n"
                    elif (
                        terms == [] and abs(intercepts[layer][n]) > prune_threshold_bias
                    ):
                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        neuron_code += f"({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field);\n"
                    elif (
                        terms != []
                        and abs(intercepts[layer][n]) <= prune_threshold_bias
                    ):
                        neuron_code += f"{' + '.join(terms)};\n"
                    else:
                        neuron_code += "0field;\n"

                    leo_code_snippets.append(neuron_code)
            prev_neurons = [
                f"neuron_{layer+1}_{n}" for n in range(coefs[layer].shape[1])
            ]

        num_outputs = coefs[-1].shape[1]
        return_line = (
            indentation
            + "return ("
            + ", ".join(
                [f"output_{i}_field as {self.leo_type}" for i in range(num_outputs)]
            )
            + ");\n"
        )
        leo_code_snippets.append(return_line)
        return leo_code_snippets

    def _transpile_mlp_logic_to_leo_code_2(
        self,
        mlp,
        model_as_input,
        indentation="",
        prune_threshold_weights=0,
        prune_threshold_bias=0,
    ):
        # optimizations through subtracting positive integer fields instead of adding negative integer fields
        leo_code_snippets = []

        coefs = mlp.coefs_
        intercepts = mlp.intercepts_

        # specify and convert inputs to fields
        for i in range(coefs[0].shape[0]):
            used_input = self.input_generator.use_input(i)

            leo_code_snippets.append(
                indentation + f"let {used_input.reference_name}_field: field = "
            )

            leo_code_snippets.append(used_input)

            leo_code_snippets.append(" as field;" + "\n")

            used_input.field_name = f"{used_input.reference_name}_field"

        # for each layer
        prev_neurons = [
            f"{self.input_generator.use_input(i).field_name}"
            for i in range(coefs[0].shape[0])
        ]
        for layer in range(len(coefs)):  # for each layer
            layer_code = []
            for n in range(coefs[layer].shape[1]):  # for each neuron in the layer
                terms = []
                for i in range(coefs[layer].shape[0]):  # for each input to the neuron
                    weight_input = self.input_generator.add_input(
                        self.leo_type,
                        "customi",
                        model_as_input,
                        coefs[layer][i][n],
                        f"w_{layer}_{n}_",
                    )
                    # todo adapt for case where model weights are actual inputs
                    if abs(coefs[layer][i][n]) > prune_threshold_weights:
                        terms.append(
                            f"({self._convert_to_fixed_point(weight_input.value.item())}{self.leo_type} as field)*{prev_neurons[i]}"
                        )

                if layer != len(coefs) - 1:  # if not the last layer
                    neuron_code = indentation + f"let neuron_{layer+1}_{n}: field = "
                    weights_or_bias_above_prune_thresholds = (
                        terms != [] or abs(intercepts[layer][n]) > prune_threshold_bias
                    )

                    if weights_or_bias_above_prune_thresholds:
                        neuron_code += "relu("  # + " + ".join(terms)
                        for i, term in enumerate(terms):
                            if i == 0:
                                if term[1] == "-":
                                    neuron_code += f" ({term[2:]}"
                                else:
                                    neuron_code += f"{term}"
                            else:
                                if term[1] == "-":
                                    neuron_code += f" - ({term[2:]}"
                                else:
                                    neuron_code += f" + {term}"

                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            neuron_code += f" + ({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field)"  # todo handle negative bias separately

                        neuron_code += ");\n"
                    else:
                        neuron_code += "0field;\n"
                    leo_code_snippets.append(neuron_code)

                else:  # if the last layer
                    neuron_code = indentation + f"let output_{n}_field" + f" : field = "
                    self.output_fixed_point_scaling_factor_power = layer + 2

                    if terms != [] and abs(intercepts[layer][n]) > prune_threshold_bias:
                        # neuron_code += f"{' + '.join(terms)}"
                        for i, term in enumerate(terms):
                            if i == 0:
                                if term[1] == "-":
                                    neuron_code += f" ({term[2:]}"
                                else:
                                    neuron_code += f"{term}"
                            else:
                                if term[1] == "-":
                                    neuron_code += f" - ({term[2:]}"
                                else:
                                    neuron_code += f" + {term}"

                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            bias_input = self.input_generator.add_input(
                                self.leo_type,
                                "customi",
                                model_as_input,
                                intercepts[layer][n],
                                f"b_{layer}_{n}_",
                            )
                            neuron_code += f" + ({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field)"
                        neuron_code += ";\n"
                    elif (
                        terms == [] and abs(intercepts[layer][n]) > prune_threshold_bias
                    ):
                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        neuron_code += f"({self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type} as field);\n"
                    elif (
                        terms != []
                        and abs(intercepts[layer][n]) <= prune_threshold_bias
                    ):
                        # neuron_code += f"{' + '.join(terms)};\n"
                        for i, term in enumerate(terms):
                            if i == 0:
                                if term[1] == "-":
                                    neuron_code += f" ({term[2:]}"
                                else:
                                    neuron_code += f"{term}"
                            else:
                                if term[1] == "-":
                                    neuron_code += f" - ({term[2:]}"
                                else:
                                    neuron_code += f" + {term}"
                        neuron_code += ";\n"
                    else:
                        neuron_code += "0field;\n"

                    leo_code_snippets.append(neuron_code)
            prev_neurons = [
                f"neuron_{layer+1}_{n}" for n in range(coefs[layer].shape[1])
            ]

        num_outputs = coefs[-1].shape[1]
        return_line = (
            indentation
            + "return ("
            + ", ".join(
                [f"output_{i}_field as {self.leo_type}" for i in range(num_outputs)]
            )
            + ");\n"
        )
        leo_code_snippets.append(return_line)
        return leo_code_snippets

    def _transpile_mlp_logic_to_leo_code_3(
        self,
        mlp,
        model_as_input,
        indentation="",
        prune_threshold_weights=0,
        prune_threshold_bias=0,
    ):
        # no usage of fields, but of layer-wise fixed point scaling factors
        leo_code_snippets = []

        coefs = mlp.coefs_
        intercepts = mlp.intercepts_

        # for each layer
        prev_neurons = [
            self.input_generator.use_input(i)
            for i in range(len(self.input_generator.input_list))
        ]
        for layer in range(len(coefs)):  # for each layer
            for n in range(coefs[layer].shape[1]):  # for each neuron in the layer
                terms = []
                for i in range(coefs[layer].shape[0]):  # for each input to the neuron
                    weight_input = self.input_generator.add_input(
                        self.leo_type,
                        "customi",
                        model_as_input,
                        coefs[layer][i][n],
                        f"w_{layer}_{n}_",
                    )
                    # todo adapt for case where model weights are actual inputs
                    if abs(coefs[layer][i][n]) > prune_threshold_weights:
                        terms.append(
                            f"{self._convert_to_fixed_point(weight_input.value.item())}{self.leo_type} * "
                        )
                        terms.append(prev_neurons[i])

                if layer != len(coefs) - 1:  # if not the last layer
                    leo_code_snippets.append(
                        indentation + f"let neuron_{layer+1}_{n}: {self.leo_type} = "
                    )
                    weights_or_bias_above_prune_thresholds = (
                        terms != [] or abs(intercepts[layer][n]) > prune_threshold_bias
                    )

                    if weights_or_bias_above_prune_thresholds:
                        leo_code_snippets.append("relu(")

                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 2 == 0:
                                leo_code_snippets.append(" + ")
                            leo_code_snippets.append(term)

                        # delete the last item if it is a plus sign
                        if leo_code_snippets[-1] == " + ":
                            leo_code_snippets.pop()

                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            leo_code_snippets.append(
                                f" + {self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type}"
                            )

                        leo_code_snippets.append(");\n")
                    else:
                        leo_code_snippets.append(f"0{self.leo_type};\n")

                else:  # if the last layer
                    neuron_code = (
                        indentation + f"let output_{n}" + f" : {self.leo_type} = "
                    )
                    self.output_fixed_point_scaling_factor_power = layer + 2

                    if terms != [] and abs(intercepts[layer][n]) > prune_threshold_bias:
                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 2 == 0:
                                neuron_code += " + "
                            neuron_code += term
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            bias_input = self.input_generator.add_input(
                                self.leo_type,
                                "customi",
                                model_as_input,
                                intercepts[layer][n],
                                f"b_{layer}_{n}_",
                            )
                            neuron_code += f" + {self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type}"
                        neuron_code += ";\n"
                    elif (
                        terms == [] and abs(intercepts[layer][n]) > prune_threshold_bias
                    ):
                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        neuron_code += f"{self._convert_to_fixed_point(bias_input.value.item(), layer+2)}{self.leo_type};\n"
                    elif (
                        terms != []
                        and abs(intercepts[layer][n]) <= prune_threshold_bias
                    ):
                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 2 == 0:
                                neuron_code += " + "
                            neuron_code += term
                        neuron_code += ";\n"
                    else:
                        neuron_code += f"0{self.leo_type};\n"

                    leo_code_snippets.append(neuron_code)
            prev_neurons = [
                f"neuron_{layer+1}_{n}" for n in range(coefs[layer].shape[1])
            ]

        num_outputs = coefs[-1].shape[1]
        return_line = (
            indentation
            + "return ("
            + ", ".join([f"output_{i}" for i in range(num_outputs)])
            + ");\n"
        )
        leo_code_snippets.append(return_line)
        return leo_code_snippets

    def _transpile_mlp_logic_to_leo_code_4(
        self,
        mlp,
        model_as_input,
        indentation="",
        prune_threshold_weights=0,
        prune_threshold_bias=0,
    ):
        # no usage of fields, and also no usage of layer-wise fixed point scaling factors
        leo_code_snippets = []

        coefs = mlp.coefs_
        intercepts = mlp.intercepts_

        # for each layer
        prev_neurons = [
            self.input_generator.use_input(i)
            for i in range(len(self.input_generator.input_list))
        ]
        for layer in range(len(coefs)):  # for each layer
            for n in range(coefs[layer].shape[1]):  # for each neuron in the layer
                terms = []
                for i in range(coefs[layer].shape[0]):  # for each input to the neuron
                    weight_input = self.input_generator.add_input(
                        self.leo_type,
                        "customi",
                        model_as_input,
                        coefs[layer][i][n],
                        f"w_{layer}_{n}_",
                    )
                    # todo adapt for case where model weights are actual inputs
                    if abs(coefs[layer][i][n]) > prune_threshold_weights:
                        terms.append(
                            f"{self._convert_to_fixed_point(weight_input.value.item())}{self.leo_type} * "
                        )
                        terms.append(prev_neurons[i])
                        terms.append(
                            f" / {self.fixed_point_scaling_factor}{self.leo_type}"
                        )

                if layer != len(coefs) - 1:  # if not the last layer
                    leo_code_snippets.append(
                        indentation + f"let neuron_{layer+1}_{n}: {self.leo_type} = "
                    )
                    weights_or_bias_above_prune_thresholds = (
                        terms != [] or abs(intercepts[layer][n]) > prune_threshold_bias
                    )

                    if weights_or_bias_above_prune_thresholds:
                        leo_code_snippets.append("relu(")

                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 3 == 0:
                                leo_code_snippets.append(" + ")
                            leo_code_snippets.append(term)

                        # delete the last item if it is a plus sign
                        if leo_code_snippets[-1] == " + ":
                            leo_code_snippets.pop()

                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            leo_code_snippets.append(
                                f" + {self._convert_to_fixed_point(bias_input.value.item())}{self.leo_type}"
                            )

                        leo_code_snippets.append(");\n")
                    else:
                        leo_code_snippets.append(f"0{self.leo_type};\n")

                else:  # if the last layer
                    neuron_code = (
                        indentation + f"let output_{n}" + f" : {self.leo_type} = "
                    )

                    if terms != [] and abs(intercepts[layer][n]) > prune_threshold_bias:
                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 3 == 0:
                                neuron_code += " + "
                            neuron_code += term
                        if abs(intercepts[layer][n]) > prune_threshold_bias:
                            bias_input = self.input_generator.add_input(
                                self.leo_type,
                                "customi",
                                model_as_input,
                                intercepts[layer][n],
                                f"b_{layer}_{n}_",
                            )
                            neuron_code += f" + {self._convert_to_fixed_point(bias_input.value.item())}{self.leo_type}"
                        neuron_code += ";\n"
                    elif (
                        terms == [] and abs(intercepts[layer][n]) > prune_threshold_bias
                    ):
                        bias_input = self.input_generator.add_input(
                            self.leo_type,
                            "customi",
                            model_as_input,
                            intercepts[layer][n],
                            f"b_{layer}_{n}_",
                        )
                        neuron_code += f"{self._convert_to_fixed_point(bias_input.value.item())}{self.leo_type};\n"
                    elif (
                        terms != []
                        and abs(intercepts[layer][n]) <= prune_threshold_bias
                    ):
                        for i, term in enumerate(terms):
                            # if i greater 0 add plus sign
                            if i > 0 and i % 3 == 0:
                                neuron_code += " + "
                            neuron_code += term
                        neuron_code += ";\n"
                    else:
                        neuron_code += f"0{self.leo_type};\n"

                    leo_code_snippets.append(neuron_code)
            prev_neurons = [
                f"neuron_{layer+1}_{n}" for n in range(coefs[layer].shape[1])
            ]

        num_outputs = coefs[-1].shape[1]
        return_line = (
            indentation
            + "return ("
            + ", ".join([f"output_{i}" for i in range(num_outputs)])
            + ");\n"
        )
        leo_code_snippets.append(return_line)
        return leo_code_snippets

    def mlp_to_pseudocode(self, mlp):
        coefs = mlp.coefs_
        intercepts = mlp.intercepts_

        code = []
        code.append(
            "function neural_network("
            + ", ".join([f"input_{i}" for i in range(coefs[0].shape[0])])
            + "):"
        )

        # for each layer
        prev_neurons = [f"input_{i}" for i in range(coefs[0].shape[0])]
        for layer in range(len(coefs)):
            layer_code = []
            for n in range(coefs[layer].shape[1]):
                terms = [
                    f"{coefs[layer][i][n]:.5f}*{prev_neurons[i]}"
                    for i in range(coefs[layer].shape[0])
                ]
                if layer != len(coefs) - 1:  # if not the last layer
                    neuron_name = f"neuron_{layer+1}_{n}"
                    layer_code.append(
                        f"    {neuron_name} = max(0, {' + '.join(terms)}"
                        f" + {intercepts[layer][n]:.5f})"
                    )
                else:  # if the last layer
                    neuron_name = f"output_{n}"
                    layer_code.append(
                        f"    {neuron_name} = {' + '.join(terms)} + "
                        f"{intercepts[layer][n]:.5f}"
                    )
            code.extend(layer_code)
            prev_neurons = [
                f"neuron_{layer+1}_{n}" for n in range(coefs[layer].shape[1])
            ]

        outputs = [f"output_{i}" for i in range(coefs[-1].shape[1])]
        code.append(f"    return softmax([{', '.join(outputs)}])")
        return "\n".join(code)



# -*- coding: utf-8 -*-
"""
LeoTranspiler: Transpile a Machine Learning model to Leo.

This package provides functionalities for transpiling a Machine Learning model to Leo.
It works with scikit-learn models.
"""
from .leo_transpiler import LeoTranspiler  # noqa: F401



# -*- coding: utf-8 -*-
def _get_rounding_decimal_places(value, max_places=10):
    if value == int(value):
        return 0
    for i in range(1, max_places + 1):
        rounded_val = round(value, i)
        if abs(value - rounded_val) < 10 ** (-i - 1):
            return i
    return max_places



# -*- coding: utf-8 -*-
import pandas as pd


class _InputGenerator:
    class _Struct:
        MAX_STRUCT_FIELDS = 32
        MAX_STRUCT_HIERARCHY = 32

        def __init__(self, inputs, horizontal_position, hierarchy, parent_struct=None):
            self.horizontal_position = horizontal_position
            self.hierarchy = hierarchy
            self.parent_struct = parent_struct
            self.leo_type = None
            self.internal_reference_name = None

            self.name = f"struct{self.hierarchy}_{self.horizontal_position}"
            if parent_struct is not None:
                self.reference_name = f"{parent_struct.reference_name}.{self.name}"
            else:
                self.reference_name = self.name

            num_inputs = len(inputs)
            if (
                num_inputs > self.MAX_STRUCT_FIELDS
                and hierarchy < self.MAX_STRUCT_HIERARCHY
            ):
                # split inputs into MAX_STRUCT_FIELDS chunks

                chunk_size = num_inputs // self.MAX_STRUCT_FIELDS
                remainder = num_inputs % self.MAX_STRUCT_FIELDS

                input_chunks = []
                start_index = 0

                # Create MAX_STRUCT_FIELDS chunks
                for i in range(self.MAX_STRUCT_FIELDS):
                    end_index = start_index + chunk_size + (1 if i < remainder else 0)
                    input_chunks.append(inputs[start_index:end_index])
                    start_index = end_index

                # Create a _Struct for each chunk, with incremented hierarchy level
                self.fields = [
                    self.__class__(
                        chunk, i, hierarchy=self.hierarchy + 1, parent_struct=self
                    )
                    for i, chunk in enumerate(input_chunks)
                ]
            elif (
                num_inputs <= self.MAX_STRUCT_FIELDS
                and hierarchy < self.MAX_STRUCT_HIERARCHY
            ):
                self.fields = inputs
                for i, field in enumerate(self.fields):
                    field.horizontal_position = i
                    field.parent_struct = self
            else:
                raise ValueError("Invalid number of inputs")

            if isinstance(self.fields[0], _InputGenerator._Input):
                for i, field in enumerate(self.fields):
                    field.name = f"x{i}"
                    field.reference_name = f"{self.reference_name}.{field.name}"

        def add_struct_definition_to_directory(self, unique_struct_directory):
            self.type_identifier = f"{self.hierarchy}"
            for field in self.fields:
                self.type_identifier += field.leo_type

            if self.type_identifier in unique_struct_directory:
                self.leo_type, _ = unique_struct_directory[self.type_identifier]
            else:
                num_structs = len(unique_struct_directory.keys())
                self.leo_type = f"Struct{num_structs}"
                if isinstance(self.fields[0], _InputGenerator._Struct):
                    leo_struct_definition = (
                        f"struct {self.leo_type}"
                        + " {\n"
                        + "  "
                        + ",\n  ".join(
                            [f"{field.name}: {field.leo_type}" for field in self.fields]
                        )
                        + "\n}"
                    )
                if isinstance(self.fields[0], _InputGenerator._Input):
                    leo_struct_definition = (
                        f"struct {self.leo_type}"
                        + " {\n"
                        + "  "
                        + ",\n  ".join(
                            [
                                f"x{i}: {field.leo_type}"
                                for i, field in enumerate(self.fields)
                            ]
                        )
                        + "\n}"
                    )
                unique_struct_directory[self.type_identifier] = (
                    self.leo_type,
                    leo_struct_definition,
                )

        def get_input_value_string(self):
            str = "{ "
            for field in self.fields:
                str += f"{field.name}: {field.get_input_value_string()}, "
            return str[:-2] + " }"

    class _Input:
        def __init__(self, value, leo_type, active, name):
            self.value = value
            self.leo_type = leo_type
            self.active = active
            self.name = name
            self.reference_name = name
            self.tmp_data_value = None
            self.horizontal_position = None
            self.parent_struct = None

        def get_set_value(self):
            if self.value is None:
                return self.tmp_data_value
            else:
                return self.value

        def get_input_value_string(self):
            return f"{self.get_set_value()}{self.leo_type}"

    def __init__(self):
        self.MAX_CIRCUIT_INPUTS = 16

        self.MAX_INPUT_VALUES = self.MAX_CIRCUIT_INPUTS * (
            self._Struct.MAX_STRUCT_FIELDS**self._Struct.MAX_STRUCT_HIERARCHY
        )
        self.input_list = []
        self._input_counts = {}

    def add_input(
        self, leo_type, naming_strategy="xi", active=False, value=None, name=None
    ):  # Todo implement custom naming strategy for different names
        input_count = self._input_counts.get(naming_strategy, 0)
        if naming_strategy == "xi":
            name = f"x{input_count}"
        elif naming_strategy == "custom":
            if name is None:
                raise Exception("Custom naming strategy requires a name")
        elif naming_strategy == "customi":
            if name is None:
                raise Exception("Custom naming strategy requires a name")
            name = f"{name}{input_count}"
        else:
            raise Exception("Invalid naming strategy")

        self._input_counts[naming_strategy] = input_count + 1

        new_input = self._Input(value, leo_type, active, name)
        self.input_list.append(new_input)

        return new_input

    def use_input(self, index):
        self.input_list[index].active = True
        return self.input_list[index]

    def get_struct_definitions_and_circuit_input_string(self):
        active_input_count = self._assign_inputs_to_structs()
        struct_definitions = self.generate_struct_definitions()
        circuit_input_string = self.generate_circuit_input_string()
        return struct_definitions, circuit_input_string, active_input_count

    def _assign_inputs_to_structs(self):
        active_inputs = [input for input in self.input_list if input.active]
        active_input_count = len(active_inputs)
        if active_input_count > self.MAX_INPUT_VALUES:
            raise ValueError(
                f"Number of active inputs ({active_input_count}) exceeds maximum "
                f"number of inputs ({self.MAX_INPUT_VALUES})"
            )

        self.structured_inputs = []
        if active_input_count <= self.MAX_CIRCUIT_INPUTS:
            self.structured_inputs = active_inputs
            for i, input in enumerate(self.structured_inputs):
                input.horizontal_position = i
                input.parent_struct = None
        else:
            # split active_inputs into MAX_CIRCUIT_INPUTS sized chunks
            # Calculate the number of elements in each chunk
            chunk_size = len(active_inputs) // self.MAX_CIRCUIT_INPUTS
            remainder = len(active_inputs) % self.MAX_CIRCUIT_INPUTS

            input_chunks = []
            start_index = 0

            # Create MAX_CIRCUIT_INPUTS chunks
            for i in range(self.MAX_CIRCUIT_INPUTS):
                end_index = start_index + chunk_size + (1 if i < remainder else 0)
                input_chunks.append(active_inputs[start_index:end_index])
                start_index = end_index

            # Create a struct for each chunk
            for i, chunk in enumerate(input_chunks):
                self.structured_inputs.append(
                    self._Struct(
                        chunk, horizontal_position=i, hierarchy=0, parent_struct=None
                    )
                )

        return active_input_count

    def generate_struct_definitions(self):
        self.unique_struct_directory = {}

        if isinstance(self.structured_inputs[0], self._Input):
            return ""

        for structured_input in self.structured_inputs:
            item = structured_input
            while True:
                if isinstance(item, self._Struct):
                    # dive deep leftmost in self.structured_inputs until hitting input
                    while isinstance(item, self._Struct):
                        item = item.fields[0]
                    item = item.parent_struct.parent_struct

                if item is not None:
                    for field in item.fields:
                        field.add_struct_definition_to_directory(
                            self.unique_struct_directory
                        )

                if item is None or item.parent_struct is None:
                    structured_input.add_struct_definition_to_directory(
                        self.unique_struct_directory
                    )
                    break
                else:
                    item = item.parent_struct

        return "\n".join(
            [
                struct_definition
                for _, struct_definition in self.unique_struct_directory.values()
            ]
        )

    def generate_circuit_input_string(self):
        input_string = ""
        for structured_input in self.structured_inputs:
            input_string += f"{structured_input.name}: {structured_input.leo_type}, "
        return input_string[:-2]

    def generate_input(self, fixed_point_features):
        inputs_without_value = len(
            [input for input in self.input_list if input.value is None]
        )
        if len(fixed_point_features) != inputs_without_value:
            raise Exception(
                f"Number of features ({len(fixed_point_features)}) does not match "
                "number of inputs without a specified value ({inputs_without_value})"
            )

        # Assign values to inputs without a specified value
        index = 0
        for input in self.input_list:
            if input.value is None:
                if isinstance(fixed_point_features, pd.Series):
                    input.tmp_data_value = fixed_point_features.iloc[index]
                else:
                    input.tmp_data_value = fixed_point_features[index]
                index += 1

        formatted_input_list = [
            el.get_input_value_string() for el in self.structured_inputs
        ]
        return formatted_input_list



# -*- coding: utf-8 -*-
"""Provides the LeoTranspiler class for transpiling ML models.

This module enables transpilation of machine learning models into the Leo
programming language, suitable for creating zero-knowledge proofs.
"""
import json
import logging
import os
import subprocess
import time
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pandas as pd
import psutil
from numpy import ndarray
from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator

from ._computation_base import LeoComputation, ZeroKnowledgeProof
from ._model_transpiler import _get_model_transpiler


class LeoTranspiler:
    """Main class for transpiling machine learning models into the Leo language.

    LeoTranspiler takes a machine learning model, and potentially some validation data,
    and provides methods to transpile this model into a Leo program. This transpiled
    program can be run to generate zero-knowledge proofs for given input samples.
    """

    def __init__(
        self,
        model: BaseEstimator,
        validation_data: Optional[ArrayLike] = None,
        output_model_hash: Optional[str] = None,
    ):
        """Initialize the LeoTranspiler with the given parameters.

        Parameters
        ----------
        model : BaseEstimator
            The ML model to transpile.
        validation_data : tuple of array_like, optional
            Data to evaluate the numerical stability of the circuit. The model will
            not be trained on this data.
        output_model_hash : str, optional
            If provided, the circuit returns the hash of the model's weights and
            biases. This functionality is not yet implemented.
        """
        self.model = model
        self.validation_data = validation_data
        self.output_model_hash = output_model_hash

        self.model_as_input = None
        self.transpilation_result = None
        self.leo_program_stored = False

    def to_leo(self, path: Path, project_name: str, model_as_input: bool = False, fixed_point_scaling_factor: Optional[int] = None, **kwargs):
        """Transpile and store the Leo program to a specified directory.

        This method transpiles the model to a Leo program and saves it, along with
        related configuration files, to the specified path under a directory named
        after the project name.

        Parameters
        ----------
        path : Path
            The directory where the Leo program and related files will be stored.
        project_name : str
            The name of the project, which determines the directory name under
            the specified path and is used during transpilation.
        model_as_input : bool, optional
            If True, the model's weights and biases are treated as circuit input
            rather than being hardcoded.

        Returns
        -------
        None

        Notes
        -----
        The method saves three main files:
        - `main.leo`: The transpiled Leo program.
        - `program.json`: Configuration for the Leo program.
        - `.env`: Environment-specific settings.

        If the Leo program is already transpiled, this method will not re-transpile
        but will directly store the program.

        Directories are created as needed to ensure the specified path exists.
        """
        self._check_installed_leo_version()
        self.model_transpiler = _get_model_transpiler(self.model, self.validation_data, fixed_point_scaling_factor, **kwargs)

        if self.transpilation_result is None or self.model_as_input != model_as_input:
            # Computing the number ranges and the fixed-point scaling factor
            logging.info("Computing number ranges and fixed-point scaling factor...")
            self.model_transpiler._numbers_get_leo_type_and_fixed_point_scaling_factor()
            logging.info("Transpiling model...")
            self.transpilation_result = self.model_transpiler.transpile(
                project_name,
                model_as_input=model_as_input,
            )  # todo check case when project name changes
            self.model_as_input = model_as_input

        self.project_dir = os.path.join(path, project_name)
        self._store_leo_program()
        self._store_program_json(project_name)
        self._store_environment_file()  # todo implement option to pass private key

        self.leo_program_stored = True
        logging.info("Leo program stored")

    def run(
        self, input: Union[ndarray, List[float], pd.DataFrame, pd.Series]
    ) -> LeoComputation:
        """Run the model to get a Leo computation for a given input sample.

        Parameters
        ----------
        input : Union[ndarray, List[float], DataFrame, Series]
            The input sample or dataset for which to generate the
            Leo computation output. Can be a numpy ndarray, a
            list of floats, a pandas DataFrame, or a pandas
            Series.

        Returns
        -------
        LeoComputation
            The Leo computation for the given input sample.
        """
        leo_computation = self._handle_input(input, "run")

        return leo_computation

    def execute(
        self, input: Union[ndarray, List[float], pd.DataFrame, pd.Series]
    ) -> ZeroKnowledgeProof:
        """Run the model to get a zero-knowledge proof for a given input sample.

        Parameters
        ----------
        input : Union[ndarray, List[float], DataFrame, Series]
            The input sample or dataset for which to generate the zero-knowledge proof.
            Can be a numpy ndarray, a list of floats, a pandas DataFrame, or a pandas
            Series.

        Returns
        -------
        ZeroKnowledgeProof
            The zero-knowledge proof for the given input sample.
        """
        zkp = self._handle_input(input, "execute")

        return zkp

    def _handle_input(self, input, command):
        if not self.leo_program_stored:
            raise FileNotFoundError("Leo program not stored")

        computation_base_result = []

        if isinstance(input, pd.DataFrame):  # an entire dataset
            for _, row in input.iterrows():
                computation_base_result.append(self._handle_run_execute(row, command))
        elif isinstance(input, list):  # a list of data points
            for data_point in input:
                computation_base_result.append(
                    self._handle_run_execute(data_point, command)
                )
        elif (
            isinstance(input, ndarray)
            and self.validation_data is not None
            and input.ndim == self.validation_data.ndim
        ):
            for data_point in input:
                computation_base_result.append(
                    self._handle_run_execute(data_point, command)
                )
        elif isinstance(input, ndarray) and self.validation_data is None:
            logging.warning(
                "No validation_data passed to the transpiler, thus, no information "
                "available of dataset shape. "
                f"Passed input sample for {command} is treated as a single data point"
            )
            computation_base_result = self._handle_run_execute(input, command)
        else:  # a single data point
            computation_base_result = self._handle_run_execute(input, command)

        return computation_base_result

    def _handle_run_execute(self, input, command):
        circuit_inputs_fixed_point = self.model_transpiler.generate_input(input)
        result, runtime = self._execute_leo_cli(command, circuit_inputs_fixed_point)
        computation_base_result = self._parse_leo_output(
            command, result, circuit_inputs_fixed_point, runtime
        )
        self.model_transpiler.convert_computation_base_outputs_to_decimal(
            computation_base_result
        )

        return computation_base_result

    def _execute_leo_cli(self, command: str, inputs: List[str]) -> Tuple[str, float]:
        """Execute a Leo CLI command.

        Parameters
        ----------
        command : str
            The command to execute.
        inputs : List[str]
            The inputs to the command.

        Returns
        -------
        Tuple[str, float]
            The command output and the runtime in seconds.
        """
        directory = self.project_dir

        # Start Leo program
        start = time.time()
        process = subprocess.Popen(
            ["leo", command, "main"] + inputs,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=directory,
        )

        while process.poll() is None:
            try:
                time.sleep(0.1)
            except psutil.NoSuchProcess:
                break

        end = time.time()

        # Get the output
        stdout, stderr = process.communicate()
        result = stdout.decode() + stderr.decode()
        runtime = end - start

        return result, runtime

    def _parse_leo_output(
        self,
        command: str,
        result: str,
        input: Optional[Union[ndarray, List[float]]] = None,
        runtime: Optional[float] = None,
    ) -> Union[LeoComputation, ZeroKnowledgeProof]:
        """Parse the Leo output.

        Parameters
        ----------
        command : str
            The command that was executed.
        result : str
            The Leo output.
        input : Union[ndarray, List[float]], optional
            The input sample for which the output was computed, by default None
        runtime : float, optional
            The runtime of the Leo computation in seconds, by default None

        Returns
        -------
        Union[LeoComputation, ZeroKnowledgeProof]
            - If the command was "run", a `LeoComputation` object.
            - If the command was "execute", a `ZeroKnowledgeProof` object.

        Raises
        ------
        ValueError
            If the command is not recognized or if there was an error parsing the
            output.
        """
        outputs_fixed_point = []

        success_run = "Finished" in result and command == "run"
        success_execute = "Executed" in result and command == "execute"
        if success_run or success_execute:
            constraints = int(
                result.split("constraints")[0].split()[-1].replace(",", "")
            )
            # Output processing
            outputs_str = result.split("Output")[1]
            outputs_str = outputs_str.split("• ")
            for element in outputs_str:
                if element.startswith("\n"):
                    continue
                # check if is number
                if element[0].isdigit() or element[0] == "-":
                    element = element.split(self.model_transpiler.leo_type)[0]
                    outputs_fixed_point.append(int(element))
        else:
            logging.error(f"Error while parsing leo outputs: {result}")
            raise ValueError("Error while parsing leo outputs")

        if success_execute:
            # get index of last \n\n in result
            index = result.rfind("\n\n")
            result_content = result[:index]
            index = result_content.rfind("\n\n")
            result_content = result_content[index + 2 :]
            # parse json
            execution_data = json.loads(result_content)

        if command == "run":
            return LeoComputation(
                input,
                outputs_fixed_point,
                constraints,
                self.model_transpiler.active_input_count,
                runtime,
            )
        elif command == "execute":
            return ZeroKnowledgeProof(
                input,
                outputs_fixed_point,
                constraints,
                self.model_transpiler.active_input_count,
                runtime,
                execution_data["execution"]["proof"],
                execution_data["execution"],
            )
        else:
            raise ValueError("Unknown command")

    def _store_leo_program(self):
        """Store the Leo program.

        Parameters
        ----------
        project_name : str
            The name of the project.

        Returns
        -------
        None
        """
        folder_dir = os.path.join(self.project_dir, "src")
        # Make sure path exists
        os.makedirs(folder_dir, exist_ok=True)

        with open(os.path.join(folder_dir, "main.leo"), "w") as f:
            f.write(self.transpilation_result)

    def _store_program_json(self, project_name: str):
        """Store the program.json file.

        Parameters
        ----------
        project_name : str
            The name of the project.

        Returns
        -------
        None
        """
        content = f"""{{
    "program": "{project_name}.aleo",
    "version": "0.0.0",
    "description": "transpiler generated program",
    "license": "MIT"
}}"""
        folder_dir = os.path.join(self.project_dir)
        # Make sure path exists
        os.makedirs(folder_dir, exist_ok=True)

        with open(os.path.join(folder_dir, "program.json"), "w") as f:
            f.write(content)

    def _store_environment_file(self):
        """Store the environment configuration file.

        Returns
        -------
        None
        """
        content = """NETWORK=testnet3
PRIVATE_KEY=APrivateKey1zkpHtqVWT6fSHgUMNxsuVf7eaR6id2cj7TieKY1Z8CP5rCD
"""
        folder_dir = os.path.join(self.project_dir)
        # Make sure path exists
        os.makedirs(folder_dir, exist_ok=True)

        with open(os.path.join(folder_dir, ".env"), "w") as f:
            f.write(content)

    def _check_installed_leo_version(self):
        """Check if Leo is installed and the version is up to date.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the Leo version is not correct.
        FileNotFoundError
            If Leo is not installed.
        """
        MIN_LEO_VERSION = "1.9.3"

        def parse_version(version):
            return tuple(map(int, version.split(".")))

        try:
            version = (
                subprocess.check_output(["leo", "--version"])
                .decode()
                .strip()
                .split(" ")[1]
            )
            if parse_version(version) < parse_version(MIN_LEO_VERSION):
                raise ValueError(
                    f"Leo version must be at least {MIN_LEO_VERSION}. Please update "
                    f"Leo using `leo update`."
                )
        except FileNotFoundError:
            raise FileNotFoundError(
                "Leo not installed. Please visit "
                "https://developer.aleo.org/leo/installation/ to install Leo."
            )



# -*- coding: utf-8 -*-
class _ComputationBase:
    """
    A base class for computational structures.

    Attributes
    ----------
    - input: The input for the computation.
    - output: The output after computation.
    - circuit_constraints: Constraints for the computation.
    - active_input_count: The number of active inputs.
    - output_decimal: Decimal representation of the output.
    - runtime: The runtime of the computation in seconds.
    - fixed_point_scaling_factor: A scaling factor (defaults to 1).
    """

    def __init__(
        self,
        input,
        output,
        circuit_constraints,
        active_input_count,
        output_decimal,
        runtime,
        fixed_point_scaling_factor=1,
    ):
        self.input = input
        self.output = output
        self.circuit_constraints = circuit_constraints
        self.active_input_count = active_input_count
        self.output_decimal = output_decimal
        self.runtime = runtime
        self.fixed_point_scaling_factor = fixed_point_scaling_factor


class ZeroKnowledgeProof(_ComputationBase):
    """
    Represents a Zero Knowledge Proof based on the ComputationBase structure.

    Attributes
    ----------
    - input: The input for the computation.
    - output: The output after computation.
    - circuit_constraints: Constraints for the computation.
    - active_input_count: The number of active inputs.
    - proof: The zero-knowledge proof.
    - execution: (Optional) Details of the execution of the proof.
    - output_decimal: Decimal representation of the output.

    Inherits other attributes from _ComputationBase.
    """

    def __init__(
        self,
        input,
        output,
        circuit_constraints,
        active_input_count,
        runtime,
        proof,
        execution=None,
        output_decimal=None,
    ):
        super().__init__(
            input,
            output,
            circuit_constraints,
            active_input_count,
            output_decimal,
            runtime,
        )
        self.proof = proof
        self.execution = execution


class LeoComputation(_ComputationBase):
    """
    Represents a Leo Computation structure based on the ComputationBase.

    Inherits attributes from _ComputationBase.
    """

    def __init__(
        self,
        input,
        output,
        circuit_constraints,
        active_input_count,
        runtime,
        output_decimal=None,
    ):
        super().__init__(
            input,
            output,
            circuit_constraints,
            active_input_count,
            output_decimal,
            runtime,
        )



# -*- coding: utf-8 -*-
from typing import Optional, Union

_leo_type_bits = [2**3, 2**4, 2**5, 2**6, 2**7]


def _get_leo_integer_bits(signed: bool, value_bits: int):
    for bits in _leo_type_bits:
        # Subtract 1 bit for signed integers
        max_bits = bits - 1 if signed else bits
        if value_bits <= max_bits:
            return bits

    raise ValueError(
        f"No leo type for {'signed' if signed else 'unsigned'} value with more than "
        f"{_leo_type_bits[-1]} bits. Try quantizing the model and/or the data."
    )


def _get_leo_integer_type(signed: bool, value_bits: int):
    for bits in _leo_type_bits:
        # Subtract 1 bit for signed integers
        max_bits = bits - 1 if signed else bits
        if value_bits <= max_bits:
            return f"{'i' if signed else 'u'}{bits}"

    raise ValueError(
        f"No leo type for {'signed' if signed else 'unsigned'} value with more than "
        f"{_leo_type_bits[-1]} bits. Try quantizing the model and/or the data."
    )


class LeoInteger:
    """Leo Integer."""

    def __init__(
        self,
        signed: bool,
        bits: int,
        original_value: Union[int, float],
        fixed_point_scaling_factor: Optional[int] = None,
    ):
        """Leo type.

        Parameters
        ----------
        value : int
            The value of the Leo type.
        type : str
            The Leo type.

        Returns
        -------
        LeoType
            The Leo type.
        """
        self.original_value = original_value
        self.type = type

        self.fixed_point_value = None  # todo
        self.decimal_value = None  # todo

        self.leo_value = self._get_leo_value(
            self.fixed_point_value, signed, bits, fixed_point_scaling_factor
        )
